""" This is the screen of the module
Copyright Nanosurf AG 2021
License - MIT
"""

from PySide2 import QtWidgets
import nanosurf as nsf
import nanosurf.lib.spm.workflow.frequency_sweep as freq_sweep
from app import module_base, app_gui
from modules.frequency_sweep_module import sweep_module, sweep_settings

class ResultTableID(nsf.gui.TableEntryIDs):
    """ identifier id are used in a nsf_table widget"""
    resonance_frequency = 0
    q_factor = 1
    spring_constant = 2

ComboPlotStyles = [
    nsf.gui.NSFComboEntry(sweep_settings.PlotStyleID.Linear,'Linear'),
    nsf.gui.NSFComboEntry(sweep_settings.PlotStyleID.Logarithmic,'Logarithmic'),
]

ComboExcitationMethods = [
    nsf.gui.NSFComboEntry(sweep_settings.ExcitationMethodID.CleanDrive,'CleanDrive'),
    nsf.gui.NSFComboEntry(sweep_settings.ExcitationMethodID.PiezoDrive,'Piezo Excitation'),
]

""" some useful lists of allowed prefixes used by nsf_sci_edit widgets"""
allowed_count_units = [nsf.sci_val.up.Prefix.base]
allowed_time_units = [nsf.sci_val.up.Prefix.base, nsf.sci_val.up.Prefix.milli]
allowed_meter_units = [nsf.sci_val.up.Prefix.milli, nsf.sci_val.up.Prefix.micro, nsf.sci_val.up.Prefix.nano]
allowed_frequency_units = [nsf.sci_val.up.Prefix.mega, nsf.sci_val.up.Prefix.kilo, nsf.sci_val.up.Prefix.base]
allowed_voltage_units = [nsf.sci_val.up.Prefix.milli, nsf.sci_val.up.Prefix.micro, nsf.sci_val.up.Prefix.base]
allowed_spring_constant_units = [nsf.sci_val.up.Prefix.base]

class SweepScreen(module_base.ModuleScreen):
    def __init__(self):
        super().__init__()

    def do_setup_screen(self, module: sweep_module.FrequencySweepModule):
        """ create here your gui with all controls and their layout"""
        self.module = module

        # left layout - main controls ------------------------------------------------------------
        self.label_cantilever = QtWidgets.QLabel("Cantilever")
        self.combo_cantilever = QtWidgets.QComboBox()
        self.update_cantilever_list()

        self.combo_excitation_method = nsf.gui.NSFComboBox(ComboExcitationMethods, "Excitation Method")

        self.label_input_source = QtWidgets.QLabel("Input Source")
        self.combo_input_source = QtWidgets.QComboBox()
        self.combo_input_source.addItems(freq_sweep.FrequencySweep.input_sources_names.values())

        self.label_output_source = QtWidgets.QLabel("Output Source")
        self.combo_output_source = QtWidgets.QComboBox()
        self.combo_output_source.addItems(freq_sweep.FrequencySweep.output_names.values())

        self.label_bandwidth = QtWidgets.QLabel("Bandwidth")
        self.combo_bandwidth = QtWidgets.QComboBox()
        self.combo_bandwidth.addItems(freq_sweep.FrequencySweep.bandwidths_names.values())

        self.scival_center_frequency = nsf.gui.NSFSciEdit("Center Frequency")
        self.scival_center_frequency.set_allowed_prefix_ids(allowed_frequency_units)
        self.scival_center_frequency.set_prefix_id(nsf.sci_val.up.Prefix.auto_)
        self.scival_center_frequency.set_precision(0)
        self.scival_center_frequency.set_value_min(0)

        self.scival_frequency_range = nsf.gui.NSFSciEdit("Frequency Range")
        self.scival_frequency_range.set_allowed_prefix_ids(allowed_frequency_units)
        self.scival_frequency_range.set_prefix_id(nsf.sci_val.up.Prefix.auto_)
        self.scival_frequency_range.set_precision(0)
        self.scival_frequency_range.set_value_min(0)

        self.scival_frequency_steps = nsf.gui.NSFSciEdit("Frequency Steps")
        self.scival_frequency_steps.set_allowed_prefix_ids(allowed_frequency_units)
        self.scival_frequency_steps.set_prefix_id(nsf.sci_val.up.Prefix.auto_)
        self.scival_frequency_steps.set_precision(0)
        self.scival_frequency_steps.set_value_min(0)

        self.scival_excitation_amplitude = nsf.gui.NSFSciEdit("Excitation Amplitude")
        self.scival_excitation_amplitude.set_allowed_prefix_ids(allowed_voltage_units)
        self.scival_excitation_amplitude.set_prefix_id(nsf.sci_val.up.Prefix.auto_)
        self.scival_excitation_amplitude.set_precision(1)
        self.scival_excitation_amplitude.set_value_min(0)
        self.scival_excitation_amplitude.set_value_max(2)
        
        self.check_emit_ticks = QtWidgets.QCheckBox("Emit Ticks")
        self.button_approach = QtWidgets.QPushButton("Approach")
        self.button_start_stop = QtWidgets.QPushButton("Start Sweep")
        self.button_save = QtWidgets.QPushButton("Save")
        

        self.layout_left = QtWidgets.QVBoxLayout()
        self.layout_left.addWidget(self.label_cantilever)
        self.layout_left.addWidget(self.combo_cantilever)
        self.layout_left.addWidget(self.combo_excitation_method)
        self.layout_left.addWidget(self.label_input_source)
        self.layout_left.addWidget(self.combo_input_source)
        self.layout_left.addWidget(self.label_output_source)
        self.layout_left.addWidget(self.combo_output_source)
        self.layout_left.addWidget(self.label_bandwidth)
        self.layout_left.addWidget(self.combo_bandwidth)
        self.layout_left.addWidget(self.scival_center_frequency)
        self.layout_left.addWidget(self.scival_frequency_range)
        self.layout_left.addWidget(self.scival_frequency_steps)
        self.layout_left.addWidget(self.scival_excitation_amplitude)
        self.layout_left.addStretch()
        self.layout_left.addWidget(self.button_start_stop)
        self.layout_left.addSpacerItem(app_gui.StdVSpacer())

        # mid layout - plots and result ---------------------------------------------------------
        self.amplitude_plot = nsf.gui.NSFChart()
        self.phase_plot = nsf.gui.NSFChart()
        self.tableResults = nsf.gui.NSFNameValueTable(ResultTableID)
        self.tableResults.define_entry(ResultTableID.resonance_frequency,"Resonance Frequency")
        self.tableResults.define_entry(ResultTableID.q_factor,"Q-Factor")

        self.layout_mid  = QtWidgets.QVBoxLayout()
        self.layout_mid.addWidget(self.amplitude_plot)
        self.layout_mid.addWidget(self.phase_plot)
        self.layout_mid.addWidget(self.tableResults)
        self.layout_mid.addSpacerItem(app_gui.StdVSpacer())

        # right layout - additional user inputs
        self.combo_plot_style = nsf.gui.NSFComboBox(ComboPlotStyles, "Plot Style")
        
        self.layout_right= QtWidgets.QVBoxLayout()
        self.layout_right.addWidget(self.combo_plot_style)
        self.layout_right.addSpacerItem(app_gui.StdVSpacer())
        
        # set GUI controls
        self.screen_layout = QtWidgets.QHBoxLayout()
        # stretch only plot area and keep controls fix in size
        self.screen_layout.addLayout(self.layout_left, 0)
        self.screen_layout.addLayout(self.layout_mid, 1) 
        self.screen_layout.addLayout(self.layout_right, 0)
        
        self.setLayout(self.screen_layout)

        self.bind_gui_elements()
        self.init_plot()
        self.enter_gui_state_idle()

    def bind_gui_elements(self):
        """ connect here all gui widgets to settings of the module or any other source"""
        
        # binding ProVal to widgets ensure that they are alway in sync 
        nsf.gui.connect_to_property(self.combo_plot_style, self.module.settings.plot_style_id)
        nsf.gui.connect_to_property(self.combo_excitation_method, self.module.settings.excitation_method)
        nsf.gui.connect_to_property(self.scival_center_frequency, self.module.settings.center_frequency)
        nsf.gui.connect_to_property(self.scival_frequency_range, self.module.settings.frequency_range)
        nsf.gui.connect_to_property(self.scival_frequency_steps, self.module.settings.frequency_steps)
        nsf.gui.connect_to_property(self.scival_excitation_amplitude, self.module.settings.excitation_amplitude)
        
        # combo boxes have to be connected separately
        self.combo_cantilever.activated.connect(self.update_cantilever)
        self.combo_input_source.activated.connect(self.update_input_source)
        self.combo_output_source.activated.connect(self.update_output_source)
        self.combo_bandwidth.activated.connect(self.update_bandwidth)

        self.combo_plot_style.value_changed_event.connect(self.update_plot)

        # buttons have to be connected separately  
        self.button_start_stop.clicked.connect(self.on_button_start_stop_clicked)
        
        # listen to signals from the core module to react and update the gui
        self.module.sig_work_start_requested.connect(self.enter_gui_state_wait)
        self.module.sig_work_stop_requested.connect(self.enter_gui_state_wait)
        self.module.sig_work_active.connect(self.enter_gui_state_active)
        self.module.sig_work_done.connect(self.enter_gui_state_idle)
        self.module.sig_new_data_available.connect(self.show_new_data)
        self.module.sig_data_invalid.connect(self.set_data_invalid)

    def init_plot(self):
        self.amplitude_plot.set_title(f"Amplitude Spectrum")
        self.amplitude_plot.set_label(nsf.gui.NSFChart.Axis.bottom, "Frequency")
        self.amplitude_plot.set_unit(nsf.gui.NSFChart.Axis.bottom, "Hz")
        self.amplitude_plot.set_label(nsf.gui.NSFChart.Axis.left, "Amplitude")
        self.amplitude_plot.set_unit(nsf.gui.NSFChart.Axis.left, "V") 
        # self.amplitude_plot.plot.vb.enableAutoRange(y=True)
        # self.amplitude_plot.plot.vb.setAutoVisible(y=False)
        self.amplitude_plot.clear_plots()

        self.phase_plot.set_title(f"Phase Spectrum")
        self.phase_plot.set_label(nsf.gui.NSFChart.Axis.bottom, "Frequency")
        self.phase_plot.set_unit(nsf.gui.NSFChart.Axis.bottom, "Hz")
        self.phase_plot.set_label(nsf.gui.NSFChart.Axis.left, "Phase")
        self.phase_plot.set_unit(nsf.gui.NSFChart.Axis.left, "deg")
        # self.phase_plot.plot.vb.enableAutoRange(y=True)
        # self.phase_plot.plot.vb.setAutoVisible(y=False)
        self.phase_plot.clear_plots()

    def on_button_start_stop_clicked(self):
        if self.module.is_sweep_busy():
            self.module.stop_sweep()
            self.start_stop_button_state(wait=True)
        else:
            self.module.start_sweep()

    def enter_gui_state_wait(self):
        self.set_parameter_widget_enable_state(enabled=False)
        self.start_stop_button_state(wait=True)

    def enter_gui_state_active(self):
        self.set_parameter_widget_enable_state(enabled=False)
        self.start_stop_button_state(wait=False, stop_state=self.module.is_sweep_busy())
        self.init_plot()

    def enter_gui_state_idle(self):
        self.set_parameter_widget_enable_state(enabled=True)
        self.start_stop_button_state(wait=False, stop_state=self.module.is_sweep_busy())

    def set_parameter_widget_enable_state(self, enabled: bool = True):
        self.combo_cantilever.setEnabled(enabled)
        self.combo_excitation_method.setEnabled(enabled)
        self.combo_input_source.setEnabled(enabled)
        self.combo_output_source.setEnabled(enabled)
        self.combo_bandwidth.setEnabled(enabled)
        self.scival_center_frequency.setEnabled(enabled)
        self.scival_frequency_range.setEnabled(enabled)
        self.scival_frequency_steps.setEnabled(enabled)
        self.scival_excitation_amplitude.setEnabled(enabled)

    def start_stop_button_state(self, wait: bool = False, stop_state: bool = False):
        if wait:
            self.button_start_stop.setEnabled(False)
            self.button_start_stop.setText("Wait...")
        else:
            self.button_start_stop.setEnabled(True)
            self.button_start_stop.setText("Stop Sweep" if stop_state else "Start Sweep")

    def show_new_data(self):
        self.update_plot()
        self.update_result_table()

    def set_data_invalid(self):
        self.amplitude_plot.clear_plots()
        self.phase_plot.clear_plots()

    def update_result_table(self):
        res = self.module.get_result()
        self.tableResults.set_value(ResultTableID.resonance_frequency, res.resonance_frequency, "Hz", precision=3)
        self.tableResults.set_value(ResultTableID.q_factor, res.q_factor, "", precision=3)

    def update_plot(self):
        if (self.combo_plot_style.value() == sweep_settings.PlotStyleID.Logarithmic):
            self.amplitude_plot.plot.setLogMode(True,True)
            self.phase_plot.plot.setLogMode(True,False)
        else:
            self.amplitude_plot.plot.setLogMode(False,False)
            self.phase_plot.plot.setLogMode(False,False)
        current_measured_data = self.module.get_sweep_result()
        current_fitted_data = self.module.get_fit_result()

        self.amplitude_plot.clear_plots()
        self.phase_plot.clear_plots()
        if current_measured_data.result_ok:
            self.amplitude_plot.plot_data(x=current_measured_data.result_freq, y=current_measured_data.result_amplitude, layer_index=0)
            self.phase_plot.plot_data(x=current_measured_data.result_freq,y=current_measured_data.result_phase, layer_index=0)
        if current_fitted_data.result_ok:
            self.amplitude_plot.plot_data(x=current_fitted_data.result_fit_freq, y=current_fitted_data.result_fit_amplitude, layer_index=1)
            self.phase_plot.plot_data(x=current_fitted_data.result_fit_freq,y=current_fitted_data.result_fit_phase, layer_index=1)

    def update_cantilever_list(self):
        self.combo_cantilever.clear()
        cantilever_list = self.module.get_cantilever_list()
        self.combo_cantilever.addItems(cantilever_list)

    def update_cantilever(self):
        self.module.cantilever.name = self.combo_cantilever.currentText()
        self.module.cantilever.index = self.combo_cantilever.currentIndex()
        self.module.select_cantilever()

    def update_excitation_method(self):
        excitation_method = self.combo_excitation_method.value()
        self.module.select_excitation_method(excitation_method)

    def update_input_source(self):
        channel_name = self.combo_input_source.currentText()
        self.module.select_input_source(channel_name)

    def update_output_source(self):
        channel_name = self.combo_output_source.currentText()
        self.module.select_output_source(channel_name)

    def update_bandwidth(self):
        channel_name = self.combo_bandwidth.currentText()
        self.module.select_bandwidth(channel_name)


"""High-level programming interface example
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import matplotlib.pyplot as plt
import nanosurf as nsf

print("Connecting to a running Nanosurf Control software...")
spm = nsf.SPM() # Depending on the software version, this could be nanosurf.C3000(), or nanosurf.CoreAFM(), etc.
if spm.is_connected():
    if spm.is_scripting_enabled():

        # make a shortcut to the application object, to make subsequent code shorter:
        application = spm.application
        application.Visible = True

        spec = application.Spec
        spec.ModuleLevel = spm.SpecModuleLevel.Standart
        spec.ModulatedOutput = spm.SpecModulatedOutput.ZAxis
        spec.ActiveZController = False
        spec.EnableRelative = True
        spec.RepetitionMode = spm.SpecRepetitionMode.Position
        spec.Repetition = 1
        spec.SpecEndMode = spm.SpecEndMode.KeepLastZPos
        spec.StartOffsetMoveSpeed = 10e-6
        spec.StartOffset = 0.0
        spec.XYMoveSpeed = 100e-6

        spec.FwdModDataPoints = 512
        spec.FwdModulationMode = spm.SpecModulationMode.FixedLength
        spec.FwdModulationRange = 500e-9
        spec.FwdModulationTime = 0.5
        spec.FwdPauseMode = spm.SpecPauseMode.KeepLastZPos
        spec.FwdPauseDataPoints = 0

        spec.BwdModDataPoints = 512
        spec.BwdModulationMode = spm.SpecModulationMode.FixedLength
        spec.BwdModulationRange = -500e-9
        spec.BwdModulationTime = 0.5
        spec.BwdPauseMode = spm.SpecPauseMode.KeepLastZPos
        spec.BwdPauseDataPoints = 0

        spec.Start()
        while spec.IsMeasuring:
            pass

        # plot the result of the measurement:
        spec_line_no = 0
        plt.plot(spec.GetLine2(
            spm.SpecGroupNo.Forward, spm.SpecChannel.Deflection_or_ZCtrlIn, spec_line_no, spm.DataFilter.RAW, spm.DataConversion.Physical), 'k')
        plt.plot(spec.GetLine2(
            spm.SpecGroupNo.Backward, spm.SpecChannel.Deflection_or_ZCtrlIn, spec_line_no, spm.DataFilter.RAW, spm.DataConversion.Physical), 'k:')
        plt.show()
    else:
        print("Sorry scripting is not activated on this controller.")
del spm


""" The functional module where the functionality goes
Copyright Nanosurf AG 2021
License - MIT
"""
import numpy as np
from PySide2.QtCore import Signal

from app import module_base, app_common
from app.app_base import ApplicationBase
from modules.demo_module import settings, worker_task

class DemoModule(module_base.ModuleBase):

    sig_work_start_requested = Signal()
    sig_work_stop_requested = Signal()
    sig_work_active = Signal()
    sig_work_done = Signal()
    sig_new_data_available = Signal()
    sig_data_invalid = Signal()

    """ Initialization functions of the module """

    def __init__(self, app: ApplicationBase, gui):
        super().__init__(app, gui)
        self.app = app
        """ Prepare here module settings which are stored and loaded from file by the app framework """
        self.settings = settings.DemoSettings()
        self.result = settings.DemoResults()

    def do_start(self):
        """ This function is called once at startup of application
            Initialize here all module specific values.
        """
        self.setup_worker_thread()
        self.connect_to_properties()

    def do_stop(self):
        """ This function is called at module shutdown"""
        if self.worker_thread.is_thread_running():
            self.logger.info("Wait until worker thread has ended...")
            self.worker_thread.stop_thread(wait=True)

    def connect_to_properties(self):
        """ Connect action functions to settings 
            The connected functions are called whenever a setting is changed (e.g. by GUI elements)
        """
        self.settings.repetitions.sig_value_changed.connect(self.update_worker_parameter)
        self.settings.time_per_repetition.sig_value_changed.connect(self.update_worker_parameter)
        self.settings.send_ticks.sig_value_changed.connect(self.update_worker_parameter)

    def setup_worker_thread(self):
        """ Create the background worker task and connect to its event """
        self.worker_thread = worker_task.MyWorker(self)
        self.worker_thread.sig_worker_started.connect(self._on_sig_worker_started)
        self.worker_thread.sig_worker_finished.connect(self._on_sig_worker_finished)
        self.worker_thread.sig_new_data.connect(self._on_sig_worker_new_data)
        self.worker_thread.sig_tick.connect(self._on_sig_worker_tick)
        self.worker_thread.start_thread()

    """ Now the business logic of the module """
        
    def start_worker(self):
        if not self.worker_thread.is_worker_running():
            self.update_worker_parameter()
            self.sig_work_start_requested.emit()
            self.worker_thread.start_worker()

    def stop_worker(self):
        if self.worker_thread.is_worker_running():
            self.sig_work_stop_requested.emit()
            self.worker_thread.abort_worker(wait=True)
    
    def is_worker_busy(self) -> bool:
        return self.worker_thread.is_worker_running()

    def get_result(self) -> settings.DemoResults:
        return self.result

    def get_worker_result(self) -> worker_task.MyWorkerData:
        return self.worker_thread.get_result()

    def update_worker_parameter(self):
        self.worker_thread.par_repetition = self.settings.repetitions.value
        self.worker_thread.par_send_ticks = self.settings.send_ticks.value
        self.worker_thread.par_time_per_repetition = self.settings.time_per_repetition.value
        self.worker_thread.par_time_between_ticks = 0.1
        self.worker_thread.par_plot_func_id = self.settings.plot_func_id

    def do_analysis(self):
        data = self.worker_thread.get_result()

        if data.last_index >= 0: 
            arr = np.array(data.value)
            mean = np.mean(arr)
            self.result.mean_value = mean
            self.result.last_data = data.value[-1]
            self.result.number_of_data_points = data.last_index
            result_valid = True
        else:
            result_valid = False

        if result_valid:
            self.sig_new_data_available.emit()
        else:
            self.app.show_message(f"Could not analyze data", app_common.MsgType.Error) 
            self.sig_data_invalid.emit()

    """ worker thread state handling """

    def _on_sig_worker_started(self):
        self.app.show_message("Working ...") 
        self.logger.info("Thread started to work")
        self.sig_data_invalid.emit()
        self.sig_work_active.emit()

    def _on_sig_worker_finished(self):
        self.sig_work_done.emit()
        if not self.worker_thread.is_worker_aborted():
            self.app.show_message("Work done") 
            self.logger.info("Background worker finished ")
            self.do_analysis()
        else:
            self.app.show_message("Work aborted") 
            self.logger.info("Background worker aborted ")
 
    def _on_sig_worker_tick(self):
        self.app.show_message("Tick") 

    def _on_sig_worker_new_data(self):
        self.do_analysis()



"""High-level programming interface example

The app demonstrates how to use the spectroscopy position table to define arbitrary spectroscopy patterns
The result of the calculation is shown in the Spectroscopy panel in the main software GUI

Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import math
import nanosurf as nsf

def create_smiley(spm: nsf.Spm, smiley_radius: float, number_of_points_for_head: int):
    """ This function defines spectroscopy coordinates looking like a smiley.

        Parameters
        ----------
        spm_controller: 
            controllers scripting interface object
        smiley_radius: float
            radius of the calculated smiley in [m]
        number_of_points_for_head: int
            defines the resolution or number of spec-points calculated for the circle which defines the head 
    """
    PI = math.pi
    step = 2*math.pi/number_of_points_for_head # degrees

    # get access to spectroscopy functions
    spec = spm.application.Spec
    spec.ClearPositionList()

    # draw head circle
    for i in range(number_of_points_for_head): 
        x = smiley_radius*math.cos(i*step)
        y = smiley_radius*math.sin(i*step)
        spec.AddPosition(x, y, 0)
    
    # draw eyes
    spec.AddPosition((smiley_radius/3), (smiley_radius/3), 0)
    spec.AddPosition((-smiley_radius/3), (smiley_radius/3), 0)

    # draw mouth
    number_of_points_for_mouth = math.trunc(number_of_points_for_head/6.0)
    for i in range(0,number_of_points_for_mouth+1):
        x = 0.75*smiley_radius*math.cos(PI*1.5-i*step)
        y = 0.75*smiley_radius*math.sin(PI*1.5-i*step)+0.1*smiley_radius
        spec.AddPosition(x, y, 0)
    for i in range(1,number_of_points_for_mouth+1):
        x = 0.75*smiley_radius*math.cos(PI*1.5+i*step)
        y = 0.75*smiley_radius*math.sin(PI*1.5+i*step)+0.1*smiley_radius
        spec.AddPosition(x, y, 0)

# Connecting to a running Nanosurf Control software...
spm = nsf.SPM()  
if spm.is_connected():
    if spm.is_scripting_enabled():

        create_smiley(spm, smiley_radius=30.0e-6, number_of_points_for_head=24)    

        # now the spectroscopy can be started to 'plot' the smiley
        # this can by done by the user in the applications GUI or by
        # spec.Start() here
    else:
        print("Sorry scripting is not activated on this controller.")
del spm
print("Done")




# studio.py


import enum
import socket
import lupa
from typing import Any

import nanosurf.lib.spm.studio.wrapper as studio_wrapper

try:
    import nanosurf.lib.spm.studio.wrapper.cmd_tree_main as cmd_tree_main
    import nanosurf.lib.spm.studio.wrapper.cmd_tree_spm as cmd_tree_spm
    import nanosurf.lib.spm.studio.wrapper.cmd_tree_ctrl as cmd_tree_ctrl
except ImportError:
    pass
except AttributeError:
    pass

class ScriptContext(enum.IntEnum):
    Main = 0
    SPM  = 1
    Ctrl = 2

g_default_ip_addr = "127.0.0.1"
g_default_ip_port = 33030

class StudioScriptInterface():
    """ This class implements the communication protocol with Studio.

        Usage
        -----

        First use connect() to establish the ip-socket.
        
        Then use execute_command() to communicate with studio. 
        More details see below.


        Implementation
        --------------

        Communication is done over ip-sockets. In most cases by "local host: 127.0.0.1" and predefined standard port 33030
            
        Commands are sent to studio as strings. 
        
        All communications are prompted by a string in this format: "{status, {result}}"
        status = 0 means no error and the result contains the result value of the command. This can be any valid Lua variable including a table itself 
        status = 1 means an error occurred and the result table contains a string with the error message. 
    """
    def __init__(self):
        self.socket:socket.socket = None
        self.server_ip = "" 
        self.server_port = 0
        self._is_server_connected = False
        self._last_error = ""

    def connect(self, host: str = g_default_ip_addr, port: int = g_default_ip_port) -> bool:
        """ Establish a connection to a running instance of studio over ip-socket

        Parameters
        ----------
        host, optional
            hosts ip-address 
        port, optional
            hosts listening port number

        Returns
        -------
            True if connection could be established. Otherwise read error message in last_error property
        """
        self.server_ip = host
        self.server_port = port 
        self.receive_buffer_size = 2**14
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.socket.connect((self.server_ip, self.server_port))
            self._is_server_connected = True
            self._last_error = ""
        except socket.error as er:
            self._is_server_connected = False
            self._last_error = str(er)
       
        return self.is_connected

    def disconnect(self):
        """ Free up ip-connection to studio. """
        self.studio = None
        del self.socket
        self.socket = None

    @property
    def is_connected(self) -> bool:
        """ Return True if connection to studio host is active and communication can be done."""
        return self._is_server_connected 

    @property
    def last_error(self) -> str:
        return self._last_error

    def execute_command(self, cmd_str: str) -> str:
        """ transmit a command string to studio and wait for the response.
        The execution of the python script is blocked until the response is received from studio

        Parameters
        ----------
        cmd_str
            Any valid command string studio accepts.

        Returns
        -------
            result string in studio format. or None in case of an error. 
            Reason of error can be read from property "self.last_error"
        """
        return_val: str = None
        try:
            # send the cmd_str with length of message as header in the first four bytes
            header_len = 4
            cmd_msg = bytes(cmd_str, encoding='utf-8')
            cmd_len = len(cmd_msg).to_bytes(header_len, signed=False, byteorder='little')
            self.socket.sendall(cmd_len + cmd_msg)

            # wait for answer. It have to be a message with header of 4 bytes describing the length of the actual result string
            res_msg = bytearray()
            rec_buffer = self.socket.recv(self.receive_buffer_size)
            if len(rec_buffer) >= header_len:
                res_len = int.from_bytes(rec_buffer[:header_len], byteorder='little', signed=False)

                rec_buffer = rec_buffer[header_len:]
                res_msg += rec_buffer 
                res_len -= len(rec_buffer)

                while res_len > 0:
                    rec_buffer = self.socket.recv(self.receive_buffer_size)
                    res_msg += rec_buffer 
                    res_len -= len(rec_buffer)

                return_val = res_msg.decode(encoding='utf-8')
            else:
                self._last_error = "Did not receive header with length of buffer"

        except socket.error as er:
            self._last_error = str(er)
        return return_val

class StudioScriptContext(StudioScriptInterface):
    """ This class represents a script context provided from studio.
        It provides comfortable access to the command tree items.

        Usage
        -----

        First use connect() to establish the communication with a session context.

        The complete command tree is read from interface and compiled into a python class tree. 
        The start of the class tree is provided in variable 'self.root'
        
        Alternatively, known command strings can be provided to call(), set() or get()

        Implementation
        --------------

        The commands Studio understands, are organized as command trees in a form like "root.workflow.imaging.start()". 
        Variables are set like "root.workflow.imaging.size = value" and read by just sending the name (e.g. "root.workflow.imaging.size")
        
        Depending on session and context, the command tree vary. Only tree "root.session" is always there.
        To get to know the actual command tree read the variable "root" and you get the serialized tree back as response.

        This command tree reading is done automatically. After the context is selected by connect()
        the complete command tree is read from interface and compiled into a python class tree. 
        The start of the class tree is provided in variable 'self.root'
        
        All communications are prompted by a result table with two arguments. "{status, {result table}}"
        status = 0 means no error and the result table contains the result of the command, 
        status = 1 means an error occurred and the result table contains a string with the error message. 
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.root: studio_wrapper.CmdTreeNode = None
        self._session_id = ""
        self._context_id = ScriptContext.Main
        self._lua = lupa.LuaRuntime(unpack_returned_tuples=True)
        self._lua_deser_func = self._lua.eval("function(ser_str) local func_c, _ = load('return '..ser_str); return func_c(); end") 
        self._root_table_name = "root_table_get_result"
        self._lua_math_type_check_func = self._lua.eval(f"function(var_name) local func_c, _ = load('return tostring(math.type({self._root_table_name}[2][1].' .. var_name .. '))'); return func_c(); end") 
                
    def connect(self, session: str, context: ScriptContext, host: str = g_default_ip_addr, port: int = g_default_ip_port) -> bool:
        """ Establish a connection to specified session and context.
            
            The complete command tree is read from interface and compiled into a python class tree. 
            The start of the class tree is provided in variable 'self.root'

        Parameters
        ----------
        session:
            name of session to open. Typically a string with serial number of controller. e.g "91-21-004" or simulated controller "91-01-000"
        context:
            context id of to open, Context "Main" is always there. Normally also context "SPM". Others are future extensions

        host, optional
            hosts ip-address, by default local host 
        port, optional
            hosts listening port number, by default standard port number

        Returns
        -------
            True if connection could be established. Otherwise error message in 'last_error' property
        """
        res = super().connect(host, port)
        if res:
            res &= self._activate_context(session, context)
        return res

    def call(self, cmd_str: str, *args, convert_result_to_python_type: bool = True) -> Any:
        """ Calls a function with arbitrary number of arguments and a return value
        
        Parameters
        ----------
        cmd_str:
            command from command tree as string. e.g. "root.workflow.imaging.start"
        args:
            argument list as normal python values. They are converted to string representations for transmitting to studio

        convert_result_to_python_type, optional:
            If set to True then the result is converted to python types. Otherwise the result may be a Lua_type. e.g. lua-table. 

        Returns
        -------
            Type depending on command. Most cases either None, float, str or list. 
            In case an error occurred, property "last_error" is not "" and contains an error description
        """
        self._last_error = ""
        arg_str = self._args_to_string(*args)
        ok, cmd_result = self._process_command(f"return {cmd_str}({arg_str})")
        if ok:
            if lupa.lua_type(cmd_result) == "table" and convert_result_to_python_type:
                cmd_result = list(cmd_result.values())
        if not ok:
            print(self._last_error)
        return cmd_result

    def get(self, lua_str: str, convert_result_to_python_type: bool = True) -> Any:
        """ Reads the current value of variable in the command tree.
        
        Parameters
        ----------
        lua_str:
            string of the command tree variable to read. e.g. "root.workflow.imaging.size"

        convert_result_to_python_type, optional:
            If set to True then the result is converted to python types. Otherwise the result may be a Lua_type. e.g. lua-table. 

        Returns
        -------
            Type depending on variable. Most cases either None, float, str or list. 
            In case an error occurred, property "last_error" is not "" and contains an error description
        """
        self._last_error = ""
        ok, cmd_result = self._process_command(f"return {lua_str}")
        if ok:
            if lupa.lua_type(cmd_result) == "table" and convert_result_to_python_type:
                cmd_result = list(cmd_result.values())
        if not ok:
            print(self._last_error)
        return cmd_result

    def set(self, lua_str: str, arg: Any) -> bool:
        """ set variable in the command tree to a new value.
        
        Parameters
        ----------
        lua_str:
            string of the command tree variable to set. e.g. "root.workflow.imaging.size"
        arg:
            new value to set as normal python type. It is converted to string representations for transmitting to studio

        Returns
        -------
            ok: 
                returns True if succeeded to set value, otherwise False and property "last_error" contains an error description
        """     
        ok = False   
        self._last_error = ""
        if lupa.lua_type(arg) != "table":
            if type(arg) == str:
                arg = self._args_to_string(arg)
            elif type(arg) == list:
                arg:str = str(arg)
                arg = '{' + arg.removeprefix('[').removesuffix(']') + '}'
            ok, _ = self._process_command(f"{lua_str}={arg}; return 0")
        else:
            ok = False
            self._last_error = "Error: Cannot set lua table directly. It have to be serialized first."
        
        if not ok:
            print(self._last_error)
        return ok

    def load_studio_root_table(self, lua_str: str) -> Any:
        res_ok = False
        res_val: dict = None
        return_val = self.execute_command(f"return {lua_str}")
        if return_val is not None:
            try:
                self._lua.execute(f"{self._root_table_name} = {return_val}")

                lua_deser_table = self._lua_deser_func(return_val)
                lua_ok = lua_deser_table[1] == 0
                res_val = lua_deser_table[2][1]
                if not lua_ok:
                    self._last_error = f"Scripting: {str(res_val)}"
                    res_val = None
                
                # re = self._lua_math_type_check_func("test.num_f")

                res_ok = True
            except lupa._lupa.LuaError as er:
                self._last_error = f"Lupa Error: {er}:\n{return_val}"
            except lupa._lupa.LuaSyntaxError:
                self._last_error = f"Lupa SyntaxError:\n{return_val}"
            except TypeError:
                self._last_error = f"Lupa TypeError:\n{return_val}"
        return res_val

    def lua_type(self, obj) -> str:
        """ convenient function to evaluate the type of a result variable. 
            If the obj provided is a lua type, then the type name is returned. Otherwise None.
            Details see lupa package documentation for lua_type()
        """
        return lupa.lua_type(obj)

    def lua_number_type_str(self, value_name : str) -> str:
        if "root." in value_name:
            value_name = value_name.removeprefix("root.")
        num_type_str = self._lua_math_type_check_func(value_name)
        return num_type_str

    def _process_command(self, cmd_str: str) -> tuple[bool, Any]:
        res_ok = False
        res_val: dict = None
        return_val = self.execute_command(cmd_str)
        if return_val is not None:
            try:
                lua_deser_table = self._lua_deser_func(return_val)
                lua_ok = lua_deser_table[1] == 0
                res_val = lua_deser_table[2][1]
                if not lua_ok:
                    self._last_error = f"Scripting: {str(res_val)}"
                    res_val = None
                res_ok = lua_ok
            except lupa._lupa.LuaError as er:
                self._last_error = f"Lupa Error: {er}:\n{return_val}"
            except lupa._lupa.LuaSyntaxError:
                self._last_error = f"Lupa SyntaxError:\n{return_val}"
            except TypeError:
                self._last_error = f"Lupa TypeError:\n{return_val}"
        return (res_ok, res_val)

    def _python_to_lua_str(self, val:str) -> str:
        cmd_str = val
        cmd_str = cmd_str.replace('\\', '\\\\')
        cmd_str = cmd_str.replace('"', '\\"')
        cmd_str = cmd_str.replace("'", "\\'")
        return cmd_str

    def _args_to_string(self, *args) -> str:
        res = ""
        for a in args:
            if type(a) == str:
                res += "'" + self._python_to_lua_str(a) + "',"
            elif type(a) == list:
                res += '{' + str(a).removeprefix('[').removesuffix(']') + '},'
            elif isinstance(a, studio_wrapper.CmdTreeNode):
                res += a._lua_tree_name + ","
            else:
                res += str(a) + ","
        return res.removesuffix(",")

    def _activate_context(self, session:str, context: ScriptContext) -> bool:
        ok = True
        self._session_id = session
        self._context_id = context
        
        if context == ScriptContext.Main:
            self.call("root.session.select_main", 0)
            ok &= (self.last_error == "") 
        else:
            self.call("root.session.select", self._session_id, context.value)
            ok &= (self.last_error == "") 

        if ok:
            ok &= self._init_cmd_tree()
        return ok

    def _init_cmd_tree(self) -> bool:
        compiler = studio_wrapper.CmdTreeCompiler(self)
        root_table = self.load_studio_root_table("root")
        ok = compiler.build_wrapper_class(self._context_id.name.lower(), root_table)
        if ok:
            if self._context_id == ScriptContext.Main:
                try:
                    import nanosurf.lib.spm.studio.wrapper.cmd_tree_main as cmd_tree_main
                    self.root = cmd_tree_main.Root(self)
                except ImportError:
                    ok = False
            elif self._context_id == ScriptContext.SPM:
                try:
                    import nanosurf.lib.spm.studio.wrapper.cmd_tree_spm as cmd_tree_spm
                    self.root = cmd_tree_spm.Root(self)
                except ImportError:
                    ok = False
            elif self._context_id == ScriptContext.Ctrl:
                try:
                    import nanosurf.lib.spm.studio.wrapper.cmd_tree_ctrl as cmd_tree_ctrl
                    self.root = cmd_tree_ctrl.Root(self)
                except ImportError:
                    ok = False
            else:
                ok = False
                self.root = None
        return ok

class StudioScriptSession():
    """ This class represents a single studio script session.
        It provides comfortable access to the command tree items of each context.
        Auto detection of active session is provided and auto activation of all contexts.

        Usage
        -----

        First use connect() to establish the communication with a session.
        If no session is provided, it try to auto-detect the active session

        For all available contexts, it creates a member variable of type StudioScriptContext

        Implementation
        --------------

        All Lua-script context of a sessions are created and stored in a variable self._context
    """  
    def __init__(self):
        # initialize all contexts and its short cuts
        self._context: dict[ScriptContext, StudioScriptContext] = {c: None for c in ScriptContext}
        self.main: cmd_tree_main.Root = None
        self.spm: cmd_tree_spm.Root = None
        self.ctrl: cmd_tree_ctrl.Root = None

        # socket interface
        self.server_ip = ""
        self.server_port = 0 
        self.session_id = ""

    def connect(self, session: str = "", host: str = g_default_ip_addr, port: int = g_default_ip_port) -> bool:
        """ Establish a connection to a running instance of studio over ip-socket

        Parameters
        ----------
        session, optional:
            name of session to open, by default auto selection of session is used
        host, optional
            hosts ip-address, by default local host 
        port, optional
            hosts listening port number, by default standard port number

        Returns
        -------
            True if connection could be established. Otherwise error message in 'last_error' property
        """
        self.server_ip = host
        self.server_port = port 
        self.session_id = session
        self._last_error = ""

        # clear all context
        self._context: dict[ScriptContext,StudioScriptContext] = {c: None for c in ScriptContext}

        # setup all context
        self._context[ScriptContext.Main] = self.create_context(ScriptContext.Main)
        if self._context[ScriptContext.Main] is not None:

            if self.session_id == "":
                self.session_id = self.auto_select_session()

            if self.session_id != "":
                self._context[ScriptContext.SPM] = self.create_context(ScriptContext.SPM)
                #self._context[ScriptContext.Ctrl] = self.create_context(ScriptContext.Ctrl) # Controller Context not yet available
            else:
                self._last_error = "No session active"

        # assign short cuts to context command trees
        self.main = self._context[ScriptContext.Main].root if self._context[ScriptContext.Main] is not None else None
        self.spm  = self._context[ScriptContext.SPM].root  if self._context[ScriptContext.SPM]  is not None else None
        self.ctrl = self._context[ScriptContext.Ctrl].root if self._context[ScriptContext.Ctrl] is not None else None

        return (self.main is not None) and (self.spm is not None)

    def auto_select_session(self) -> str:
        session_list = self.get_sessions()
        return session_list.pop() if len(session_list) > 0 else ""

    def create_context(self, context: ScriptContext) -> StudioScriptContext:
        new_context = StudioScriptContext()
        ok = new_context.connect(self.session_id, context, self.server_ip, self.server_port)
        if not(ok):
            self._last_error = f"Connect to session '{self.session_id}' context '{context}' failed.\nReason: '{new_context.last_error}'"
            new_context = None
        return new_context

    def get_sessions(self, session: str = None) -> list[str]:
        sessions = []
        if self._context[ScriptContext.Main] is not None:
            sessions = self._context[ScriptContext.Main].root.session.list()
        return sessions

    def disconnect(self):
        """ Close all session nodes"""
        for val in self._context.values():
            if val is not None:
                val.disconnect()

    @property
    def is_connected(self) -> bool:
        """ Return True if connection to studio host is active and communication can be done."""
        return self.main is not None

    @property
    def last_error(self) -> str:
        return self._last_error



"""Package for scripting the Nanosurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import time
import logging
import enum
import typing
import nanosurf.lib.spm as spm
import nanosurf


class MotorID():
    """ Defines the identification parameters of a direct motor control motor.
        Has to be constructed with initial parameters for each motor. Normally the scan head class is doing so.
    """
    def __init__(self, proxy_index: int, lu_index: int, range: float):
        self.ProxyIndex = proxy_index
        self.LUIndex = lu_index
        self.Range = range

class MoveStatus(enum.IntEnum):
    """ Defines motor move status"""
    Undefined = 0,
    IdleUnreferenced = 1,
    IdleReferenced = 2,
    Referencing = 3,
    Moving = 4

class LimitStatus(enum.IntEnum):
    """ Defines motor limit switch status"""
    Undefined = 0
    InNoLimit = 1,
    InUpperLimit = 2,
    InLowerLimit = 3,
    InBothLimits = 4,
    InMultipleLimits = 5

class Direction(enum.IntEnum):
    """ Defines motor move directions for various functions """
    Forward = 0,
    Backward = 1

class MotorState():
    def __init__(self):
        # Axis Status
        self.move_state = MoveStatus.Undefined
        self.limit_state = LimitStatus.Undefined

class StudioMotorControl:
    """ Base class for dealing with Nanosurf Motors. In this base class general motor control functions are used.
        For measurement head specific functions, see the corresponding child classes
    """

    class _MoveMode(enum.IntEnum):
        Absolute = 0,
        Relative = 1,

    def __init__(self, spm: spm.Spm = None, *args):
        """Base class for dealing with Nanosurf Motors. In this base class general motor control functions are used.
           For measurement head specific functions, see the corresponding child classes
        """
        self.connected = False
        self.spm = None
        self.spm_app = None
        self.spm_system = None
        self.lu_direct_motor_ctrl = None
        self.logger = logging.getLogger("Motor_Control")
        self.tick_callback = self.default_tick_callback

        self.__sleep_time_after_motor_command = 0.2 # seconds
        self.__motor_speed = 1.0 # full speed
        self.__status_poll_time = 0.0
        self.__max_wait_time = 20.0 # [s]

        self.map_move_status = [
            MoveStatus.IdleUnreferenced,
            MoveStatus.IdleReferenced,
            MoveStatus.Referencing,
            MoveStatus.Moving
        ]
        self.map_limit_status = [
            LimitStatus.InNoLimit,
            LimitStatus.InUpperLimit,
            LimitStatus.InLowerLimit,
            LimitStatus.InBothLimits,
            LimitStatus.InMultipleLimits
        ]

        self.studio = nanosurf.Studio()
        self.studio.connect()
        self.lu_direct_motor_ctrl = self.studio.spm.lu.DirectMotorControl.Instance
        self.lu_system = self.studio.spm.lu.System.Instance

    @property
    def motor_speed(self) -> float:
        """ returns the current set motor speed used far all motor movements"""
        return self.__motor_speed

    @motor_speed.setter
    def motor_speed(self, speed: float):
        """ Set all motor speeds.
        Parameter
        ----------
            val: float
                speed values range is: 0.01 > value <=1.0.
                1.0 means full speed
        """
        self.__motor_speed = min(max(0.01, speed), 1.0)

    @property
    def sleep_time_after_motor_command(self):
        """ Current sleep time in [s] after each COM motor command.  """
        return self.__sleep_time_after_motor_command

    @sleep_time_after_motor_command.setter
    def sleep_time_after_motor_command(self, val):
        """ Sleep time in [s] after each COM motor command.
            It already happened, when two motor move command to separate axes were sent through the COM interface in a row,
            there was a race condition and the same axis got twice the same command.
            Therefore it is suggested to set this parameter to 200ms, or to zero; when you don't want to wait.
        """
        self.__sleep_time_after_motor_command = min(max(0.0, val), 1.0)

    @property
    def status_poll_time(self):
        """ Sleep time in [s] during status polling. """
        return self.__status_poll_time

    @status_poll_time.setter
    def status_poll_time(self, val):
        """ Sleep time in [s] during status polling.
        """
        self.__status_poll_time = min(max(0.0, val), 1.0)

    @property
    def max_wait_time(self):
        """ Maximal time in [s] to wait for end of movement is reached. """
        return self.__max_wait_time

    @max_wait_time.setter
    def max_wait_time(self, val):
        """ Maximal time in [s] to wait for end of movement is reached.
        """
        self.__max_wait_time = min(max(0.0, val), 120.0)

    def get_motor_status(self, motor_id: MotorID) -> MotorState:
        """ collects current axis states for a motor"""
        status = MotorState()
        self.lu_direct_motor_ctrl.Attribute.CurrentAxis.SetValue(motor_id.LUIndex)
        axis_status = self.lu_direct_motor_ctrl.Attribute.AxisStatus.GetValue()
        limit_switch_status = self.lu_direct_motor_ctrl.Attribute.AxisLimitStatus.GetValue()
        status.move_state = self.map_move_status[int(axis_status)]
        status.limit_state = self.map_limit_status[int(limit_switch_status)]
        return status

    def start_move_relative(self, motor_id: typing.Union[MotorID, list[MotorID]], step_size: typing.Union[float, list[float]]):
        """ Start a relative move from current position

        Paramters
        ---------
        motor_id:  single motorID or list of motorIDs
        step_size: float or list of float
            travel length. can be positive or negative

        """
        self.logger.debug("start motor(s) relative move")
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        self._move_command(motor_list, step_size, StudioMotorControl._MoveMode.Relative, wait_for_end=False)

    def do_move_relative(self, motor_id: typing.Union[MotorID, list[MotorID]], step_size: typing.Union[float, list[float]]) -> bool:
        """ start and wait a relative move from current position

        Paramters
        ---------
        motor_id:  single motorID or list of motorIDs
        step_size: float or list of float
            travel length. can be positive or negative

        """
        self.logger.debug("do motor(s) relative move")
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        return self._move_command(motor_list, step_size, StudioMotorControl._MoveMode.Relative, wait_for_end=True)

    def start_move_absolute(self, motor_id: typing.Union[MotorID, list[MotorID]], target_position: typing.Union[float, list[float]]):
        """ Start a move to absolute target position(s)

        Paramters
        ---------
        motor_id:  single motorID or list of motorIDs
        target_position: float or list of float
            travel length. can be positive or negative

        """
        self.logger.debug("start motor(s) move to absolute position")
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        self._move_command(motor_list, target_position, StudioMotorControl._MoveMode.Absolute, wait_for_end=False)

    def do_move_absolute(self, motor_id: typing.Union[MotorID, list[MotorID]], target_position: typing.Union[float, list[float]]) -> bool:
        """ start and wait a move to absolute target position(s)

        Paramters
        ---------
        motor_id:  single motorID or list of motorIDs
        target_position: float or list of float
            target position has to be positive

        """
        self.logger.debug("do motor(s) move to absolute position")
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        return self._move_command(motor_list, target_position, StudioMotorControl._MoveMode.Absolute, wait_for_end=True)

    def overwrite_max_axis_speed(self, motor_id: typing.Union[MotorID, list[MotorID]], new_max_speed: float):
        """ update the max speed of motors defined in the hardware.
        WARNING:
        --------
        Please never change the maximum motor speed, without safety checks.
        It can damage the hardware.
        The max speed has no upper limit as it overwrites the software defined dmc max speed
        """
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        for index, motor in enumerate(motor_list):
            self.lu_direct_motor_ctrl.Attribute.CurrentAxis.SetValue(motor.LUIndex)
            self.lu_direct_motor_ctrl.Attribute.CurrentAxis.SetValue(new_max_speed)
        time.sleep(self.sleep_time_after_motor_command)

    # ------- implementation ------------------------------------------

    def _move_command(self, motors: list[MotorID], move_positions: typing.Union[float, list[float]], move_mode: _MoveMode, wait_for_end: bool) -> bool:
        """ handle the moving command """
        self.lu_direct_motor_ctrl.Trigger.ResetTransaction()
        for index, motor in enumerate(motors):
            move_pos = move_positions[index] if isinstance(move_positions, list) else move_positions
            self.lu_direct_motor_ctrl.Attribute.CurrentAxis.SetValue(motor.LUIndex)
            self.lu_direct_motor_ctrl.Attribute.AxisSpeedFactor.SetValue(self._get_lu_motor_speed())
            self.lu_direct_motor_ctrl.Attribute.TransactionMoveRelative.SetValue(move_mode.value)
            self.lu_direct_motor_ctrl.Attribute.TransactionMoveDestination.SetValue(move_pos)
            self.lu_direct_motor_ctrl.Trigger.AddTransactionMove()

        self.lu_direct_motor_ctrl.Trigger.CommitTransaction()
        time.sleep(self.sleep_time_after_motor_command)

        if wait_for_end:
            return self._wait_for_end_of_movement(motors)
        return True

    def _wait_for_end_of_movement(self,  motors: list[MotorID]) -> bool:
        """ Wait until all motors are stopped or a timeout is reached.

        Paramters
        ---------
        motors: list of motorIDs

        Returns
        -------
         done: bool
            return True if all defined motors are stopped in timout time. False if Timeout was reached
        """
        motor_moving = True
        time_out_reached = False
        start_time = time.time()
        already_checked_once = False
        while motor_moving and not time_out_reached and not already_checked_once:
            stopped = [] # list of stop state for each motor
            for index, motor in enumerate(motors):
                status = self.get_motor_status(motor)
                if (status.move_state == MoveStatus.IdleReferenced) or (status.move_state == MoveStatus.IdleUnreferenced):
                    stopped.append(True)
                else:
                    stopped.append(False)

            if all(stopped) and already_checked_once:
                motor_moving = False
                self.logger.debug("All motor stopped")
            elif(all(stopped) and not already_checked_once):
                already_checked_once = True # We check once more
                self.logger.debug("Check again whether motors stopped")
            else:
                if (time.time() - start_time) > self.max_wait_time:
                    time_out_reached = True
                    self.logger.debug("Motors where still moving after timeout reached")

            self.tick_callback(time.time() - start_time)
            time.sleep(self.status_poll_time)
        return not time_out_reached

    def _get_lu_motor_speed(self) -> float:
        """ for LU it does not use speed factor between 0 and 1 but between 0 and 100"""
        return self.motor_speed * 100.0

    def default_tick_callback(self, time: float):
        pass


""" Line chart and color mpa Widgets based on pyqtgraph library
Copyright Nanosurf AG 2021
License - MIT
"""

from __future__ import print_function
import enum
import typing
import numpy as np
from PySide2 import QtGui
import pyqtgraph as pg
import pyqtgraph.functions as fn
import nanosurf.lib.datatypes.sci_channel as sci_channel
import nanosurf.lib.datatypes.sci_stream as sci_stream
from nanosurf.lib.gui import nsf_colors

pg.setConfigOption('foreground', 'w')
pg.setConfigOption('background', 'k')
pen1D = pg.mkPen(color='w', width=2)
color_nsf_second_line = nsf_colors.NSFColorsTuple.Soft_White
color_nsf_main_line = nsf_colors.NSFColorsTuple.Orange

class NSFChart(pg.GraphicsLayoutWidget):

    class Axis(enum.Enum):
        bottom = "bottom"
        left = "left"
        right = "right"
        top = "top"

    class AxisText:
        def __init__(self, name: str = "", unit: str = ""):
            self.name = name
            self.unit = unit

    
    def __init__(self, title: str = "Chart Plot", logmodex=False, logmodey=False, labelsize=12, titelsize=14):
        pg.GraphicsLayoutWidget.__init__(self)
        self.labelsize = labelsize
        self.titlesize = titelsize
        self.logmode_x = logmodex 
        self.logmode_y = logmodey
        self.axis_text = {axisname: NSFChart.AxisText() for axisname, _ in NSFChart.Axis.__members__.items()} 
        
        font = QtGui.QFont()
        font.setPixelSize(self.labelsize )

        self.plot = self.addPlot()
        self.plot.setLogMode(self.logmode_x,self.logmode_y)
        self.plot.getAxis(NSFChart.Axis.bottom.value).tickFont = font
        self.plot.getAxis(NSFChart.Axis.bottom.value).setStyle(tickTextOffset = 3, tickLength = 5)
        self.plot.getAxis(NSFChart.Axis.left.value).tickFont = font
        self.plot.getAxis(NSFChart.Axis.left.value).setStyle(tickLength = 5)

        self.set_label(NSFChart.Axis.bottom, "X-Axis")
        self.set_label(NSFChart.Axis.left, "Y-Axis")
        self.set_title(title)

        self.plot_layers = [None, None]
        self.plot_layers[1] = self.plot.plot(pen = color_nsf_second_line)
        self.plot_layers[0] = self.plot.plot(pen = color_nsf_main_line) 
                                              
        self.view_box = self.plot.vb
        self.view_box.setMouseMode(1) # single mouse button mode
        self.plot.scene().sigMouseMoved.connect(self._move_cross_hair_cursor)
        self.plot.scene().leaveEvent = self._mouse_left_event_handler
        self.setMinimumHeight(200)
        self.setMinimumWidth(200)

        self._add_cross_hair_cursor()
        self._create_marker()
        self.clear_plots()

    def set_title(self, title: str):
        self.plot_title = title
        self._show_title(self.plot_title)

    def get_title(self) -> str:
        return self.plot_title

    def set_label(self, axis: Axis, label: str):
        self.axis_text[axis.value].name = label
        self._show_label(axis)

    def set_unit(self, axis: Axis, unit: str):
        self.axis_text[axis.value].unit = unit
        self._show_label(axis)

    def plot_data(self, y: typing.Union[list, np.ndarray], x: typing.Union[list, np.ndarray, None] = None, layer_index: int = 0):
        
        if isinstance(x, np.ndarray):
            x = x.tolist()
        if isinstance(y, np.ndarray):
            y = y.tolist()
            
        if layer_index < len( self.plot_layers):
            p = self.plot_layers[layer_index]

            if x is not None:
                x_data = x     
            else:
                x_data = [x for x in range(len(y))]
            p.setData(x=x_data, y=y)

    def plot_channel(self, channel:sci_channel.SciChannel, layer_index: int = 0):
        if layer_index < len( self.plot_layers):
            p = self.plot_layers[layer_index]
            self.set_unit(NSFChart.Axis.left, channel.unit) 
            p.setData(y=channel.value) 

    def plot_stream(self, stream: sci_stream.SciStream, channel_index: int = 0, layer_index: int = 0):
        if layer_index < len( self.plot_layers):
            p = self.plot_layers[layer_index]
            ch = stream.get_channel(channel_index)
            self.set_unit(NSFChart.Axis.bottom, stream.x.unit) 
            self.set_unit(NSFChart.Axis.left, ch.unit) 
            p.setData(x=stream.x.value, y=ch.value) 

    def extend_plot_layers(self, num_layers: int):
        if num_layers > len(self.plot_layers):
            missing_layers = num_layers - len(self.plot_layers)
            self.plot_layers.extend([self.plot.plot(pen = color_nsf_second_line) for i in range(missing_layers)])

    def clear_plots(self):
        for p in self.plot_layers:
            p.setData(x=[], y=[])    

    def get_plot_layer(self, layer_index: int): 
        if layer_index < len(self.plot_layers): 
            return  self.plot_layers[layer_index]
        return None

    def get_plot_layer_count(self) -> int:
        return len(self.plot_layers) 
        
    def set_marker(self, x: typing.Union[float, list[float]], y: typing.Union[float, list[float]]):
        if isinstance(x, float):
            self.plt_marker.setData(x=[x],y=[y])            
        else:
            self.plt_marker.setData(x=x,y=y)            

    def clear_marker(self):
        self.plt_marker.setData(x=[],y=[])   

    # implementation -------------------------------------------------

    def _show_title(self, title: str):
        self.plot.setTitle(title=title, size=f"{self.titlesize}px")

    def _show_label(self, axis:Axis):
        self.plot.setLabel(axis.value, text=f'<span style="font-size:{self.labelsize}pt">{self.axis_text[axis.value].name}</span>', units=self.axis_text[axis.value].unit)

    def _create_marker(self):
        self.plt_marker = self.plot.plot(pen=color_nsf_second_line, symbol='+', symbolSize=15, ) 

    def _add_cross_hair_cursor(self):
        self.cursor_vLine = pg.InfiniteLine(angle=90, movable=False)
        self.cursor_hLine = pg.InfiniteLine(angle=0, movable=False)
        self.plot.addItem(self.cursor_vLine, ignoreBounds=True)
        self.plot.addItem(self.cursor_hLine, ignoreBounds=True)
        self.cursor_vLine.setPos(-1000)
        self.cursor_hLine.setPos(-1000)
        
    def _move_cross_hair_cursor(self, event):
        pos = event
        if self.plot.sceneBoundingRect().contains(pos):
            mousePoint = self.view_box.mapSceneToView(pos)
            x = mousePoint.x()
            y = mousePoint.y()
            self.cursor_vLine.setPos(x)
            self.cursor_hLine.setPos(y)

            if self.logmode_x:
                x = 10**x
            if self.logmode_y:
                y = 10**y
            x_unit =self.plot.getAxis(NSFChart.Axis.bottom.value).labelUnits
            y_unit =self.plot.getAxis(NSFChart.Axis.left.value).labelUnits
            self._show_title(f"x={fn.siFormat(x,suffix=x_unit, precision=3)} y={fn.siFormat(y,suffix=y_unit, precision=3)}")

    def _mouse_left_event_handler(self, event):
        self._show_title(self.plot_title)
        self._clear_cross_hair_cursor()

    def _clear_cross_hair_cursor(self):
        self.cursor_vLine.setPos(-1000)
        self.cursor_hLine.setPos(-1000)
                           

class NSFColormap(pg.GraphicsLayoutWidget):

    class Axis(enum.Enum):
        bottom = "bottom"
        left = "left"
        right = "right"
        top = "top"
        z = "z"

    class AxisText:
        def __init__(self, name: str = "", unit: str = ""):
            self.name = name
            self.unit = unit

    def __init__(self, title: str = "Colormap Plot", logmodex: bool = False, logmodey: bool = False, labelsize: int = 14, titlesize: int = 16):
        pg.GraphicsLayoutWidget.__init__(self)
        self.labelsize = labelsize
        self.titlesize = titlesize
        self.logmode_x = logmodex 
        self.logmode_y = logmodey  
        self.axis_text = {axisname: NSFColormap.AxisText() for axisname, _ in NSFColormap.Axis.__members__.items()} 

        self.datapoints = (0,0)
        self.datamatrix = np.array([], ndmin=2)    
 
        font = QtGui.QFont()
        font.setPixelSize(self.labelsize)

        self.plot = self.addPlot()
        self.plot.setLogMode(self.logmode_x, self.logmode_y)
        self.plot.getAxis(NSFColormap.Axis.bottom.value).tickFont = font
        self.plot.getAxis(NSFColormap.Axis.bottom.value).setStyle(tickTextOffset = 3, tickLength = 5)
        self.plot.getAxis(NSFColormap.Axis.left.value).tickFont = font
        self.plot.getAxis(NSFColormap.Axis.left.value).setStyle(tickLength = 5)

        self.colormap = pg.ImageItem()
        self.plot.addItem(self.colormap)
        self.plot.invertY(False)

        self.set_title(title)
        self._add_cross_hair_cursor()
        self.colormap.hoverEvent = self._move_cross_hairs
        self.set_label(NSFColormap.Axis.bottom, "X-Axis")
        self.set_label(NSFColormap.Axis.left, "Y-Axis")
        self._show_title(self.plot_title)
        self.histogram = pg.HistogramLUTWidget()
        self.histogram.setImageItem(self.colormap)
        self.histogram.gradient.loadPreset('spectrum')
        self.histogram.hide()

        self.setMinimumHeight(200)
        self.setMinimumWidth(200)

    def set_label(self, axis: Axis, label: str):
        self.axis_text[axis.value].name = label
        self._show_label(axis)

    def set_unit(self, axis: Axis, unit: str):
        self.axis_text[axis.value].unit = unit
        self._show_label(axis)

    def set_title(self, title: str):
        self.plot_title = title
        self._show_title(self.plot_title)

    def set_data_points(self, x: int, y: int):
        """ Defines the number of data points of the 2D matrix
            With this the matrix of data to show is defined.
            Update of the matrix is done by the plot_xxxxx() functions
        """
        self.datapoints = (x,y)
        xx , yy = np.meshgrid(np.zeros(x), np.zeros(y), indexing='ij')
        self.datamatrix = xx

    def set_xy_range(self, x: np.ndarray, y:np.ndarray):
        """ Defines the x/y range for the matrix in physical units

        Parameters
        ----------
        x, y: ndarray
            in these arrays the range of each axis is defined.
            The min and max values are used to define the left/right or bottom/top values of the axes
            So, the array content can be just two values [min, max] or a full stream of data of any length
        """
        # prepare ImageItem transformation:
        tr = QtGui.QTransform()  
        tr.rotate(90.0)
        tr.scale((np.max(x) - np.min(x))/len(x),-1.0* (np.max(y) - np.min(y))/len(y))  # scale horizontal and vertical axes
        self.colormap.setTransform(tr)

        self.plot.setXRange(np.min(x), np.max(x))
        self.plot.setYRange(np.min(y), np.max(y))
        self.colormap.setPos(np.min(x),np.min(y))

    def plot_data_point(self, x_pos: int, y_pos: int, value: float):
        """ Update a single point in the matrix at x/y
        
        Parameters
        ----------
        x,y: int
            Coordinate in the matrix of the point. 
            (0,0) is at bottom, left
            (max_x, max_y) is at top/right of the colormap
        value: float
            The value of the data point at (x,y)
        """
        self.datamatrix[y_pos][x_pos] = value
        self.colormap.setImage(self.datamatrix)

    def plot_data_line(self, data_line: np.ndarray, y_pos: int):
        """ Update a single row in the matrix at (y)
        
        Parameters
        ----------
        y_pos: int
            Y-Coordinate in the matrix of the point. 
            (0) is at bottom, (max_y) is at top of the colormap
        data_line: ndarray
            Array of values to The value of the data point at (x,y)
        """
        self.datamatrix[y_pos] = data_line
        self.colormap.setImage(self.datamatrix)

    def plot_channel(self, channel: sci_channel.SciChannel, y_pos: int):
        if y_pos >= 0 :
            self.datamatrix[y_pos] = channel.value
            self.colormap.setImage(self.datamatrix)
            self.set_unit(NSFColormap.Axis.left, channel.unit)         

    def plot_matrix(self, xy: np.ndarray):
        self.datamatrix = xy
        self.colormap.setImage(self.datamatrix)

    # Implementation ------------------------------------

    def _show_title(self, title: str):
        self.plot.setTitle(title = title, size = f"{self.titlesize}px")

    def _show_label(self, axis:Axis):
        if axis != NSFColormap.Axis.z:
            self.plot.setLabel(axis.value, text=f'<span style="font-size:{self.labelsize}pt">{self.axis_text[axis.value].name}</span>', units=self.axis_text[axis.value].unit)

    def _add_cross_hair_cursor(self):
        self.cursor_vLine = pg.InfiniteLine(angle=90, movable=False)
        self.cursor_hLine = pg.InfiniteLine(angle=0, movable=False)
        self.plot.addItem(self.cursor_vLine, ignoreBounds=True)
        self.plot.addItem(self.cursor_hLine, ignoreBounds=True)
        self.cursor_vLine.setPos(-1000)
        self.cursor_hLine.setPos(-1000)
            
    def _move_cross_hairs(self, event):
        if event.isExit():
            self._show_title(self.plot_title)
            self.cursor_vLine.setPos(-1000)
            self.cursor_hLine.setPos(-1000)
            return

        pos = event.pos()
        i, j = pos.y(), pos.x()
        i = int(np.clip(i, 0, self.colormap.image.shape[0] - 1)) # xpixel
        j = int(np.clip(j, 0, self.colormap.image.shape[1] - 1)) # ypixel
        val = self.colormap.image[j, i]
        ppos = self.colormap.mapToParent(pos)
        x, y = ppos.x(), ppos.y()
        self.cursor_vLine.setPos(x)
        self.cursor_hLine.setPos(y)

        x_unit = self.plot.getAxis(NSFColormap.Axis.bottom.value).labelUnits
        y_unit = self.plot.getAxis(NSFColormap.Axis.left.value).labelUnits
        z_unit = self.axis_text[NSFColormap.Axis.z.value].unit
        self.set_title( 
            f"x={fn.siFormat(x,suffix=x_unit, precision=3)}, y={fn.siFormat(y,suffix=y_unit, precision=3)}, z={fn.siFormat(val,suffix=z_unit,precision=3)}",\
        )




""" Defines the configuration values of the module.
    Values in the ProStore are saved and loaded during application startup/shutdown automatically
Copyright Nanosurf AG 2021
License - MIT
"""

import enum
import numpy as np
import pathlib
import nanosurf.lib.datatypes.sci_val as sci_val
import nanosurf.lib.datatypes.prop_val as prop_val

class OutputChannelID(enum.IntEnum):
    User1 = 0,
    User2 = 1,
    TipVoltage = 2


class SpecSettings(prop_val.PropStore):
    """ settings defined here as PropVal are stored persistently in a ini-file
        settings with a '_' as first char are not stored
    """
    def __init__(self):
        self.output_id = prop_val.PropVal(int(OutputChannelID.User1))
        self.number_of_steps = prop_val.PropVal(int(20))
        self.time_delay_after_step = prop_val.PropVal(sci_val.SciVal(0.05, "s"))
        self.output_span = prop_val.PropVal(sci_val.SciVal(20.0, "V"))
        self.output_center = prop_val.PropVal(sci_val.SciVal(0.0, "V"))
        self.folder_name = prop_val.PropVal(pathlib.Path(""))
        self.file_name_mask = prop_val.PropVal(str("switching-spec-"))
        self.file_index = prop_val.PropVal(int(1))
        self.auto_save_data = prop_val.PropVal(bool(True))
        self.show_on_data_values = prop_val.PropVal(bool(False))

class SpecResults():
    """ This class saves the worker task result (e.g be read by gui elements or saved to file """
    def __init__(self):
        self.outputs : np.ndarray = np.array([], dtype=float)
        self.outputs_unit: str = ""
        self.amplitudes_on : np.ndarray = np.array([], dtype=float)
        self.amplitudes_off : np.ndarray = np.array([], dtype=float)
        self.amplitudes_unit: str = ""
        self.phases_on : np.ndarray = np.array([], dtype=float)
        self.phases_off : np.ndarray = np.array([], dtype=float)




""" The functional module where the functionality goes
Copyright Nanosurf AG 2021
License - MIT
"""

import time
import nanosurf
import nanosurf.lib.spm.workflow.frequency_sweep as freq_sweep

from PySide2.QtCore import Signal

from app import module_base
from app import app_common
from app.app_base import ApplicationBase
from modules.frequency_sweep_module import sweep_settings, sweep_task, sweep_fit

class FrequencySweepModule(module_base.ModuleBase):

    sig_work_start_requested = Signal()
    sig_work_stop_requested = Signal()
    sig_work_active = Signal()
    sig_work_tick = Signal(float) # inform about sweep progress
    sig_work_done = Signal()
    sig_new_data_available = Signal()
    sig_data_invalid = Signal()

    """ Initialization functions of the module """

    def __init__(self, app: ApplicationBase, gui):
        super().__init__(app, gui)
        self.app = app
        self.spm:nanosurf.spm.Spm = None
        self.freq_sweeper:freq_sweep.FrequencySweep = None

        """ Prepare here module settings which are stored and loaded from file by the app framework """
        self.settings = sweep_settings.SweepSettings()
        self.cantilever = self.settings.cantilever
        self.result = sweep_settings.SweepResults()

    def do_start(self):
        """ This function is called once at startup of application
            Initialize here all module specific values.
        """
        if self._connect_to_controller():
            self.application = self.spm.application
            self.scan_head = self.application.ScanHead
            self.freq_sweeper = freq_sweep.FrequencySweep(self.spm)

        self.setup_sweep_thread()
        self.setup_fit_thread()
        self.connect_to_properties()

    def do_stop(self):
        """ This function is called at module shutdown"""
        if self.sweep_thread.is_thread_running():
            self.logger.info("Wait until worker thread has ended...")
            self.sweep_thread.stop_thread(wait=True)

        if self.fit_thread.is_thread_running():    
            self.fit_thread.stop_thread(wait=True)

    def get_cantilever_list(self):
        self.cantilever_list = []
        try:
            cantilever_list_count = self.application.CantileverList.Count
            for i in range(0, cantilever_list_count):
                self.cantilever_list.append(self.application.CantileverList.name(i))
        except:
            pass
        return self.cantilever_list

    def connect_to_properties(self):
        """ Connect action functions to settings 
            The connected functions are called whenever a setting is changed (e.g. by GUI elements)
        """
        self.settings.center_frequency.sig_value_changed.connect(self.update_sweep_parameter)
        self.settings.frequency_range.sig_value_changed.connect(self.update_sweep_parameter)
        self.settings.frequency_steps.sig_value_changed.connect(self.update_sweep_parameter)
        self.settings.excitation_amplitude.sig_value_changed.connect(self.update_sweep_parameter)
        self.settings.plot_style_id.sig_value_changed.connect(self.update_sweep_parameter)

    def setup_sweep_thread(self):
        """ Create the sweep worker task and connect to its event """
        self.sweep_thread = sweep_task.FrequencySweepWorker(self)
        self.sweep_thread.sig_worker_started.connect(self._on_sig_sweep_started)
        self.sweep_thread.sig_worker_finished.connect(self._on_sig_sweep_finished)
        self.sweep_thread.sig_sweep_tick.connect(self._on_sig_sweep_ticker)
        self.sweep_thread.start_thread()

    def setup_fit_thread(self):
        """ Create the fit worker task and connect to its event """
        self.fit_thread = sweep_fit.FrequencySweepFitWorker(self)
        self.fit_thread.sig_worker_started.connect(self._on_sig_fit_started)
        self.fit_thread.sig_worker_finished.connect(self._on_sig_fit_finished)
        self.fit_thread.start_thread()
    
    """ Now the business logic of the module """

    def _connect_to_controller(self) -> bool:
        ok = self.spm is not None
        if not ok:
            self.spm = nanosurf.SPM()
            if self.spm.is_connected():
                if self.spm.is_scripting_enabled():
                    ok = True
                else:
                    self.app.show_message("Error: Scripting interface is not enabled", app_common.MsgType.Error)
            else:
                 self.app.show_message("Error: Could not connect to controller. Check if software is started", app_common.MsgType.Error)
        if not ok:
            self.spm = None
        return ok

    def _disconnect_from_controller(self):
        if self.spm is not None:
            if self.spm.application is not None:
                del self.spm
            self.spm = None

    def select_cantilever(self):
        self.scan_head.Cantilever = self.cantilever.index
        self.get_cantilever_properties()

    def get_cantilever_properties(self):
        self.cantilever.length = self.scan_head.GetCantileverProperty(0)
        self.cantilever.width = self.scan_head.GetCantileverProperty(1)
        self.cantilever.spring_constant = self.scan_head.GetCantileverProperty(2)
        self.cantilever.resonance_frequency_air = self.scan_head.GetCantileverProperty(3)
        self.cantilever.q_factor_air = self.scan_head.GetCantileverProperty(4)
        self.cantilever.resonance_frequency_liquid = self.scan_head.GetCantileverProperty(5)
        self.cantilever.q_factor_liquid = self.scan_head.GetCantileverProperty(6)

    def select_excitation_method(self, excitation_method: sweep_settings.ExcitationMethodID):
        if excitation_method == sweep_settings.ExcitationMethodID.CleanDrive: 
            self.settings.excitation_method = nanosurf.Spm.ExcitationMode.PhotoThermal
        else: 
            self.settings.excitation_method = nanosurf.Spm.ExcitationMode.PiezoElectric
        self.update_sweep_parameter()

    def select_input_source(self, channel_name: dict):
        key_list = list(self.freq_sweeper.input_sources_names.keys())
        val_list = list(self.freq_sweeper.input_sources_names.values())
        try:
            pos = val_list.index(channel_name)
            self.settings.input_source = freq_sweep.InputSource[key_list[pos].name]
            self.update_sweep_parameter()
        except:
            pass      

    def select_output_source(self, channel_name: dict):
        key_list = list(freq_sweep.FrequencySweep.output_names.keys())
        val_list = list(freq_sweep.FrequencySweep.output_names.values())
        try:
            pos = val_list.index(channel_name)
            self.settings.output_source=freq_sweep.FrequencySweepOutput[key_list[pos].name]
            self.update_sweep_parameter()
        except:
            pass

    def select_bandwidth(self, channel_name: dict):
        key_list = list(self.freq_sweeper.bandwidths_names.keys())
        val_list = list(self.freq_sweeper.bandwidths_names.values())
        try:
            pos = val_list.index(channel_name)
            self.settings.bandwidth = freq_sweep.Bandwidths[key_list[pos].name]
            self.update_sweep_parameter() 
        except:
            pass

    def start_sweep(self):
        if not self.sweep_thread.is_worker_running():
            self.update_sweep_parameter()
            self.sweep_thread.start_worker()

    def stop_sweep(self):
        if self.sweep_thread.is_worker_running():
            self.app.show_message("Aborting Sweep...")
            self.sweep_thread.abort_worker(wait=False)
    
    def is_sweep_busy(self) -> bool:
        return self.sweep_thread.is_worker_running()

    def get_result(self) -> sweep_settings.SweepResults:
        return self.result

    def get_sweep_result(self) -> sweep_task.FrequencySweepData:
        return self.sweep_thread.get_result()
    
    def start_fit(self):
        if not self.fit_thread.is_worker_running():
            self.update_sweep_parameter()
            self.update_fit_parameter()
            self.fit_thread.start_worker()
    
    def get_fit_result(self) -> sweep_fit.FrequencySweepFitData:
        return self.fit_thread.get_result()

    """ Parameter update"""   

    def update_sweep_parameter(self):
        self.sweep_thread.par_cantilever = self.cantilever
        self.sweep_thread.par_input_source = self.settings.input_source
        self.sweep_thread.par_output_source = self.settings.output_source
        self.sweep_thread.par_bandwidth=self.settings.bandwidth

        self.sweep_thread.par_center_frequency = self.settings.center_frequency.value
        self.sweep_thread.par_frequency_range = self.settings.frequency_range.value
        self.sweep_thread.par_frequency_step = self.settings.frequency_steps.value
        self.sweep_thread.par_excitation_amplitude = self.settings.excitation_amplitude.value
        self.sweep_thread.par_deflection_setpoint = self.settings.deflection_setpoint.value
        self.sweep_thread.par_excitation_method = self.settings.excitation_method.value

    def update_fit_parameter(self):
        data = self.sweep_thread.get_result()
        self.fit_thread.par_measured_amplitudes = data.result_amplitude
        self.fit_thread.par_measured_phases = data.result_phase
        self.fit_thread.par_measured_frequencys = data.result_freq

    """ thread state handling """

    """ sweep thread state handling """
    def _on_sig_sweep_started(self):
        self.app.show_message("Sweeping ...") 
        self.logger.info("Thread started to work")
        self.sig_data_invalid.emit()
        self.sig_work_active.emit()

    def _on_sig_sweep_finished(self):
        if not self.sweep_thread.is_worker_aborted():
            self.app.show_message("Sweep done")
            self.sig_new_data_available.emit()            
            self.start_fit()
            self.logger.info("Thread done ")
        else:
            self.app.show_message("Sweep aborted") 
            self.logger.info("Thread aborted ")
        self.sig_work_done.emit()

    def _on_sig_sweep_ticker(self, cur_freq, remaining_time):
        self.app.show_message(f"Sweeping in progress. Current frequency: {cur_freq:.2f}Hz, Remaining time {remaining_time:.1f}s.") 
        self.sig_work_tick.emit(remaining_time)

    """ fit thread state handling """
    def _on_sig_fit_started(self):
        self.app.show_message("Fitting ...") 
        self.logger.info("Thread started to work")
        self.sig_work_active.emit()

    def _on_sig_fit_finished(self):
        if not self.sweep_thread.is_worker_aborted():
            fit_data = self.fit_thread.get_result()
            if fit_data.result_ok:
                self.app.show_message("Fitting done")
                self.logger.info("Fitting done")
                self.result.resonance_frequency = fit_data.result_resonance_frequency_amplitude
                self.result.q_factor = fit_data.result_q_factor_amplitude
            else:
                self.app.show_message("Could not extract amplitude and phase curve from data", app_common.MsgType.Error) 
                self.app.show_message("Fitting error") 
                self.logger.info("Thread aborted ")
        else:
            self.app.show_message("Fitting aborted") 
            self.logger.info("Thread aborted ")
        self.sig_new_data_available.emit()
        self.sig_work_done.emit()
